# License:
#   Copyright (C) 2011 Thomas GUILLET, Damien CHAPON, Marc LABADENS. All Rights Reserved.
#
#   This file is part of PyMSES.
#
#   PyMSES is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   PyMSES is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with PyMSES.  If not, see <http://www.gnu.org/licenses/>.
import wx
from widget_ids import WidgetIds as wid
from utils import get_icon
from axes3d import Axes3D
from glass import MagGlass
from map_tabs import GenericTabMapWindow
from pymses.utils import misc
from pymses.analysis.visualization.fft_projection.convolution_kernels import ConvolKernel

class Expander(wx.CollapsiblePane):#{{{
	def __init__(self, parent, id, title, model, controller, is_collapsed=True):
		wx.CollapsiblePane.__init__(self, parent, id, label=title, style= wx.CP_DEFAULT_STYLE | wx.CP_NO_TLW_RESIZE)

		self.model = model
		self.controller = controller
		
		pane = self.GetPane()
		sizer = wx.BoxSizer(wx.VERTICAL)
		pane.SetSizer(sizer)
		
		box = wx.StaticBox(pane, wx.ID_ANY)
		boxsizer = wx.StaticBoxSizer(box, wx.VERTICAL)
		inpane = wx.Panel(pane, wx.ID_ANY)
		self.insizer = wx.BoxSizer(wx.VERTICAL)
		inpane.SetSizer(self.insizer)
		
		self.make_content(inpane)
		
		boxsizer.Add(inpane, 0, wx.ALL|wx.EXPAND, 2)
		sizer.Add(boxsizer, 0, wx.EXPAND|wx.LEFT, 10)
		self.Collapse(is_collapsed)

	def add_descr_label(self, parent, text):
		label = wx.StaticText(parent, wx.ID_ANY, "%s :"%text)
		self.insizer.Add(label, 0, wx.ALIGN_LEFT)
		
	def add_centered_item(self, item, is_expanded=False):
		if is_expanded:
			self.insizer.Add(item, 0, wx.LEFT|wx.EXPAND, 5)
		else:
			self.insizer.Add(item, 0, wx.LEFT|wx.ALIGN_CENTER_HORIZONTAL, 5)


	def make_content(self, parent, sizer):
		raise NotImplementedError()

	def reset(self):
		raise NotImplementedError()

#}}}

class ParametersExpander(Expander):#{{{
	def __init__(self, parent, model, controller):
		WID = wid()
		Expander.__init__(self, parent, WID.EXPANDER_PARAMETERS, "Parameters", model, controller, is_collapsed=True)

	def make_content(self, parent):#{{{
		WID = wid()
		
		self.add_descr_label(parent, "Region finder map")
		p_one = wx.Panel(parent, WID.ANY)
		self.regionFinderMapType_cb = wx.ComboBox(p_one, WID.ANY, size=(120,-1), style=wx.CB_READONLY)
		self.listregionFinderMapType = ["density","temperature","velocity","Sigma","dm_particles", "stars_particles"]
		self.regionFinderMapType_cb.SetItems(self.listregionFinderMapType)
		self.regionFinderMapType_cb.SetValue("density")
		self.model.regionFinderMapType = "density"
		self.regionFinderMapType_cb.SetToolTip(wx.ToolTip("Change the region finder map type"))
		self.add_centered_item(p_one)
		
		self.add_descr_label(parent, "Multiprocessing")
		p_two = wx.Panel(parent, WID.ANY)
		bs = wx.BoxSizer(wx.HORIZONTAL)
		p_two.SetSizer(bs)
		self.Multiprocessing_cb = wx.ComboBox(p_two, WID.ANY, size=(80,-1), style=wx.CB_READONLY)
		self.listMultiprocessing_cb = ["1","2","4","8","16"]
		self.Multiprocessing_cb.SetItems(self.listMultiprocessing_cb)
		self.Multiprocessing_cb.SetValue(str(misc.NUMBER_OF_PROCESSES_LIMIT))
		self.Multiprocessing_cb.SetToolTip(wx.ToolTip("Limit the number of process created with multiprocessing"))
		self.MultiprocessingBtn = wx.lib.buttons.GenBitmapToggleButton(p_two, WID.ANY, wx.Bitmap(get_icon("memory.png")))
		self.MultiprocessingBtn.SetToolTip(wx.ToolTip("Use multiprocessing to draw the 3 region finder map in parallel, USE ONLY IF YOU HAVE ENOUGH RAM MEMORY"))
		self.MultiprocessingBtn.SetToggle(False)
		self.model.multiprocessing = self.MultiprocessingBtn.GetToggle()
		bs.Add(self.Multiprocessing_cb, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
		bs.Add(self.MultiprocessingBtn, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_LEFT, 5)
		self.add_centered_item(p_two)
		
		p_resolution = wx.Panel(parent, WID.ANY)
		bs = wx.BoxSizer(wx.HORIZONTAL)
		p_resolution.SetSizer(bs)
		p_resolution_text = wx.StaticText(p_resolution, wx.ID_ANY, "Resolution : ")
		resolution_toolTip_text = "This multiply computed images resolution by this lower resolution factor to get views faster"
		p_resolution_text.SetToolTip(wx.ToolTip(resolution_toolTip_text))
		bs.Add(p_resolution_text, 0,wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_LEFT, 5)
		self.lower_resolution_cb = wx.ComboBox(p_resolution, WID.ANY, size=(60,-1), style=wx.CB_READONLY)
		self.lower_resolution_list = ["1","0.8","0.5","0.3","0.1"]
		self.lower_resolution_cb.SetItems(self.lower_resolution_list)
		self.lower_resolution_cb.SetValue("1")
		self.model.rf_lower_resolution = None
		self.model.tab_lower_resolution = None
		self.lower_resolution_cb.SetToolTip(wx.ToolTip(resolution_toolTip_text))
		self.lower_resolution_cb.Bind(wx.EVT_COMBOBOX, self.update_lower_resolution_cb)
		bs.Add(self.lower_resolution_cb, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
		#self.add_centered_item(p_resolution)
		self.insizer.Add(p_resolution, 0, wx.LEFT, 0)
		
		self.add_descr_label(parent, "FFT kernel size factor")
		p_three = wx.Panel(parent, WID.ANY)
		bs = wx.BoxSizer(wx.HORIZONTAL)
		p_three.SetSizer(bs)
		self.FFTkernelSize_cb = wx.ComboBox(p_three, WID.ANY, size=(80,-1), style=wx.CB_READONLY)
		self.listFFTkernelSizeStringValue = ["0.5","1","2","4"]
		self.FFTkernelSize_cb.SetItems(self.listFFTkernelSizeStringValue)
		self.FFTkernelSize_cb.SetValue("1")
		self.FFTkernelSize_cb.SetToolTip(wx.ToolTip("Update FFT kernel ( = point) size factor"))
		self.model.FFTkernelSizeFactor = 1
		self.rememberSomeDataBtn = wx.lib.buttons.GenBitmapToggleButton(p_three, WID.ANY, wx.Bitmap(get_icon("memory.png")))
		self.rememberSomeDataBtn.SetToolTip(wx.ToolTip("Remember some data in cache, USE ONLY IF YOU HAVE ENOUGH RAM MEMORY, and click here again to free all memory"))
		self.rememberSomeDataBtn.SetToggle(True)
		self.model.rememberSomeData = self.rememberSomeDataBtn.GetToggle()
		bs.Add(self.FFTkernelSize_cb, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
		bs.Add(self.rememberSomeDataBtn, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_LEFT, 5)
		self.add_centered_item(p_three)
		self.random_shiftCheckBox = wx.CheckBox(parent, -1, 'random shift', (20, 100))
		self.random_shiftCheckBox.SetToolTip(wx.ToolTip("Add a random shift to point positions to avoid grid effect on FFT splatting images"))
		self.model.random_shift = False
		self.add_centered_item(self.random_shiftCheckBox)
		
		
		self.add_descr_label(parent, "Image postprocessing")
		p_four = wx.Panel(parent, WID.ANY)
		bs = wx.BoxSizer(wx.HORIZONTAL)
		p_four.SetSizer(bs)
		self.cmap_cb = wx.ComboBox(p_four, WID.ANY, size=(120,-1), style=wx.CB_READONLY)
		self.list_cmap = ["jet","Greys_r","Reds","Greens","Blues","Reds_r","Greens_r","Blues_r","Spectral","RdBu", "RdBu_r"]
		self.cmap_cb.SetItems(self.list_cmap)
		self.cmap_cb.SetValue("jet")
		self.cmap_cb.SetToolTip(wx.ToolTip("Change the colormap used"))
		self.AdaptiveGaussianBlurBtn = wx.lib.buttons.GenBitmapToggleButton(p_four, WID.ANY, wx.Bitmap(get_icon("gaussian_blur.png")))
		self.AdaptiveGaussianBlurBtn.SetToolTip(wx.ToolTip("Experimental : compute local image resolution and apply an adaptive gaussian blur"))
		self.AdaptiveGaussianBlurBtn.SetToggle(False)
		self.model.adaptive_gaussian_blur = self.AdaptiveGaussianBlurBtn.GetToggle()
		bs.Add(self.cmap_cb, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
		bs.Add(self.AdaptiveGaussianBlurBtn, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_LEFT, 5)
		self.add_centered_item(p_four)
		
		
		misc.init_OpenCl()
		self.model.use_openCL = False
		if misc.OpenCL_initialized:
			# boundary line for OpenCL
			self.insizer.Add(wx.StaticText(parent, wx.ID_ANY, "         -------------------------"), 0, wx.ALIGN_LEFT)
			p_five = wx.Panel(parent, WID.ANY)
			bs = wx.BoxSizer(wx.HORIZONTAL)
			p_five.SetSizer(bs)
			self.openCLCheckBox = wx.CheckBox(p_five, -1, 'Use OpenCL device', (40, 100))
			self.openCLCheckBox.SetToolTip(wx.ToolTip("Use OpenCL if possible: works only with ray tracing with fully loaded octreedatasource yet"))
			self.openCL_cb = wx.ComboBox(p_five, WID.ANY, size=(50,-1), style=wx.CB_READONLY)
			self.openCL_list = ["0","1","2"]
			self.openCL_cb.SetItems(self.openCL_list)
			self.openCL_cb.SetValue("0")
			bs.Add(self.openCLCheckBox, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_LEFT, 5)
			bs.Add(self.openCL_cb, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
			self.add_centered_item(p_five)
			# Bind the button
			self.openCLCheckBox.Bind(wx.EVT_CHECKBOX, self.update_openCLCheckBox)
			self.openCL_cb.Bind(wx.EVT_COMBOBOX, self.update_openCL_cb)
		
		# Bind the button
		self.FFTkernelSize_cb.Bind(wx.EVT_COMBOBOX, self.updateFFTkernelSize)
		self.random_shiftCheckBox.Bind(wx.EVT_CHECKBOX, self.updateRandom_shift)
		self.rememberSomeDataBtn.Bind(wx.EVT_BUTTON, self.rememberSomeDataMethod)
		self.Multiprocessing_cb.Bind(wx.EVT_COMBOBOX, self.updateMultiprocessingLimit)
		self.MultiprocessingBtn.Bind(wx.EVT_BUTTON, self.MultiprocessingBtnMethod)
		self.AdaptiveGaussianBlurBtn.Bind(wx.EVT_BUTTON, self.AdaptiveGaussianBlurBtnMethod)
		self.regionFinderMapType_cb.Bind(wx.EVT_COMBOBOX, self.updateRegionFinderMapType)
		self.cmap_cb.Bind(wx.EVT_COMBOBOX, self.update_cmap)
	#}}}
	
	def update_lower_resolution_cb(self, event):
		if event.GetSelection() == 0:
			self.model.rf_lower_resolution = None
			self.model.tab_lower_resolution = None
		else:
			nxy_list = self.controller.widgets[wid().RF_TAB].GetWindowSizesDict()
			self.model.update_cameras_dict(nxy_list)
			self.model.rf_lower_resolution = int(float(self.lower_resolution_list[event.GetSelection()])\
								* self.model.cam_dict['los'].map_max_size)
			self.model.tab_lower_resolution = int(self.model.rf_lower_resolution * 2.1)
		self.controller.UpdateImages()
	
	def update_openCLCheckBox(self, event):
		self.model.use_openCL = self.openCLCheckBox.GetValue()
		print "use_openCL =", self.model.use_openCL
	
	def update_openCL_cb(self, event):
		misc.init_OpenCl(event.GetSelection())
		if not self.model.use_openCL:
			self.openCLCheckBox.SetValue(True)
			self.update_openCLCheckBox(event)
			
	
	def updateRandom_shift(self, event):
		self.model.random_shift = self.random_shiftCheckBox.GetValue()
		print "random_shift =", self.model.random_shift
		
	def updateFFTkernelSize(self, event):
		self.model.FFTkernelSizeFactor = float(self.listFFTkernelSizeStringValue[event.GetSelection()])
		self.controller.UpdateImages()
	
	def rememberSomeDataMethod(self, event):
		self.rememberSomeDataBtn.SetToggle(self.rememberSomeDataBtn.GetToggle())
		self.model.rememberSomeData = self.rememberSomeDataBtn.GetToggle()
		if not self.model.rememberSomeData:
			self.model.freeTheMemory()
		print "Remember some data =", self.model.rememberSomeData
		
	def updateMultiprocessingLimit(self, event):
		misc.NUMBER_OF_PROCESSES_LIMIT =  int(self.listMultiprocessing_cb[event.GetSelection()])
		
	def MultiprocessingBtnMethod(self, event):
		self.MultiprocessingBtn.SetToggle(self.MultiprocessingBtn.GetToggle())
		self.model.multiprocessing = self.MultiprocessingBtn.GetToggle()
		print "Region Finder Multiprocessing =", self.model.multiprocessing
	
	def AdaptiveGaussianBlurBtnMethod(self, event):
		self.AdaptiveGaussianBlurBtn.SetToggle(self.AdaptiveGaussianBlurBtn.GetToggle())
		self.model.adaptive_gaussian_blur = self.AdaptiveGaussianBlurBtn.GetToggle()
		self.controller.UpdateImages(do_all_processing=False, verbose=False)
	
	def updateRegionFinderMapType(self, event):
		self.model.regionFinderMapType = self.listregionFinderMapType[event.GetSelection()]
		self.model.me = None
		print "Region finder map :", self.model.regionFinderMapType
		if self.model.regionFinderMapType == "velocity":
			self.cmap_cb.SetValue("RdBu_r")
			self.model.cmap = "RdBu_r"
		else:
			self.cmap_cb.SetValue("jet")
			self.model.cmap = "jet"
	
	def update_cmap(self, event):
		self.model.cmap = self.list_cmap[event.GetSelection()]
		self.controller.UpdateImages(do_all_processing=False, verbose=False)
		
	def reset(self):
		self.FFTkernelSize_cb.SetValue("1")
#}}}

class LineOfSightExpander(Expander):#{{{
	def __init__(self, parent, model, controller):
		WID = wid()
		Expander.__init__(self, parent, WID.EXPANDER_LOS, "Line-of-sight axis", model, controller, is_collapsed=False)

	def make_content(self, parent):#{{{
		WID = wid()
		self.add_descr_label(parent, "3D-view")
		self.a3d = Axes3D(parent)
		self.add_centered_item(self.a3d)
		self.a3d.Bind(wx.EVT_MOUSE_EVENTS, self.controller.A3dOnMouse)
		self.a3d.Bind(wx.EVT_KEY_DOWN, self.controller.A3dOnKeyPress)
		self.a3d.Bind(wx.EVT_KEY_UP, self.controller.A3dOnKeyRelease)

		self.add_descr_label(parent, "Axis shortcuts")
		axis_up_toolbar = wx.ToolBar(parent, WID.ANY, style=wx.TB_HORIZONTAL | wx.NO_BORDER)
		axis_up_toolbar.AddSimpleTool(WID.LOS_TOOL_X, wx.Bitmap(get_icon("x-axis.png")), 'View along the X axis')
		axis_up_toolbar.AddSeparator()
		axis_up_toolbar.AddSimpleTool(WID.LOS_TOOL_Y, wx.Bitmap(get_icon("y-axis.png")), 'View along the Y axis')
		axis_up_toolbar.AddSeparator()
		axis_up_toolbar.AddSimpleTool(WID.LOS_TOOL_Z, wx.Bitmap(get_icon("z-axis.png")), 'View along the Z axis')
		axis_up_toolbar.AddSeparator()
		axis_up_toolbar.AddSimpleTool(WID.LOS_TOOL_ROT_LEFT, wx.Bitmap(get_icon("rotate_left.png")), 'Rotate the axes to the left')
		axis_up_toolbar.AddSeparator()
		axis_up_toolbar.AddSimpleTool(WID.LOS_TOOL_ROT_TOP, wx.Bitmap(get_icon("rotate_top.png")), 'Rotate the axes to the top')
		axis_up_toolbar.Realize()
		axis_down_toolbar = wx.ToolBar(parent, WID.ANY, style=wx.TB_HORIZONTAL | wx.NO_BORDER)
		axis_down_toolbar.AddSimpleTool(WID.LOS_TOOL_MX, wx.Bitmap(get_icon("x-axis.png")), 'View along the -X axis')
		axis_down_toolbar.AddSeparator()
		axis_down_toolbar.AddSimpleTool(WID.LOS_TOOL_MY, wx.Bitmap(get_icon("y-axis.png")), 'View along the -Y axis')
		axis_down_toolbar.AddSeparator()
		axis_down_toolbar.AddSimpleTool(WID.LOS_TOOL_MZ, wx.Bitmap(get_icon("z-axis.png")), 'View along the -Z axis')
		axis_down_toolbar.AddSeparator()
		axis_down_toolbar.AddSimpleTool(WID.LOS_TOOL_ROT_RIGHT, wx.Bitmap(get_icon("rotate_right.png")), 'Rotate the axes to the right')
		axis_down_toolbar.AddSeparator()
		axis_down_toolbar.AddSimpleTool(WID.LOS_TOOL_ROT_BOTTOM, wx.Bitmap(get_icon("rotate_bottom.png")), 'Rotate the axes to the bottom')
		axis_down_toolbar.Realize()
		self.add_centered_item(axis_up_toolbar, is_expanded=True)
		self.add_centered_item(axis_down_toolbar, is_expanded=True)
		
		self.autorefreshCheckBox = wx.CheckBox(parent, -1, 'Auto Refresh', (20, 100))
		self.autorefreshCheckBox.SetToolTip(wx.ToolTip("Automatically update images when the mouse is released on 3D-view"))
		self.controller.A3d_autorefresh = False
		#self.add_centered_item(self.autorefreshCheckBox)
		self.insizer.Add(self.autorefreshCheckBox, 0, wx.LEFT, 0)
		
		self.add_descr_label(parent, "Line-of-sight coordinates")
		coord_panel = wx.Panel(parent, WID.ANY)
		grid_sizer = wx.FlexGridSizer(3,2)
		coord_panel.SetSizer(grid_sizer)
		los_x_label =  wx.StaticText(coord_panel, WID.ANY, "x :")
		los_y_label =  wx.StaticText(coord_panel, WID.ANY, "y :")
		los_z_label =  wx.StaticText(coord_panel, WID.ANY, "z :")
		self.los_x = wx.TextCtrl(coord_panel, WID.ANY, style=wx.TE_READONLY, size=(80,-1))
		self.los_y = wx.TextCtrl(coord_panel, WID.ANY, style=wx.TE_READONLY, size=(80,-1))
		self.los_z = wx.TextCtrl(coord_panel, WID.ANY, style=wx.TE_READONLY, size=(80,-1))
		grid_sizer.AddMany([(los_x_label, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT),\
							(self.los_x, 0), \
							(los_y_label, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT), \
							(self.los_y, 0), \
							(los_z_label, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT), \
							(self.los_z, 0)])
		self.add_centered_item(coord_panel)

		# Binding toolbar events to controller method
		axis_up_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_X)
		axis_down_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_MX)
		axis_up_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_Y)
		axis_down_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_MY)
		axis_up_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_Z)
		axis_down_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_MZ)
		axis_up_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_ROT_LEFT)
		axis_down_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_ROT_RIGHT)
		axis_up_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_ROT_TOP)
		axis_down_toolbar.Bind(wx.EVT_TOOL, self.controller.OnShortcutAxis, id=WID.LOS_TOOL_ROT_BOTTOM)
		self.autorefreshCheckBox.Bind(wx.EVT_CHECKBOX, self.updateAutorefresh)
	#}}}

	def RefreshAxes3D(self):#{{{
		"""
		Line-of-sight expander content refresh method

		mouse_in  --- whether the mouse is in the 3D window or not
		"""
		# Axes3D refresh
		self.a3d.SetUVPoints(self.model.GetUVPoints())
		self.a3d.Refresh()

		# Line-of-sight x/y/z coordinates update
		los = self.model.GetLosAxis()
		self.los_x.ChangeValue("%.6f"%los[0])
		self.los_y.ChangeValue("%.6f"%los[1])
		self.los_z.ChangeValue("%.6f"%los[2])
	#}}}

	def GetAxes3d(self):#{{{
		return self.a3d
	#}}}
	
	def updateAutorefresh(self, event):
		self.controller.A3d_autorefresh = self.autorefreshCheckBox.GetValue()

	def reset(self):#{{{
		self.a3d.SetCtrlDown(False)
		self.a3d.SetClickDown(False)
		self.a3d.SetMouseIn(False)
		self.a3d.SetCursorPosition(None, None)
		self.a3d.UpdateCursor()
		self.RefreshAxes3D()
	#}}}
#}}}

class MagGlassExpander(Expander):#{{{
	def __init__(self, parent, model, controller):
		WID = wid()
		Expander.__init__(self, parent, WID.EXPANDER_GLASS, "Magnifier", model, controller, is_collapsed=True)

	def make_content(self, parent):#{{{
		WID = wid()
		self.add_descr_label(parent, "Magnifying glass")
		self.glass = MagGlass(parent)
		self.add_centered_item(self.glass)
		
		self.add_descr_label(parent, "Value")
		self.glass_value = wx.StaticText(parent, WID.GLASS_VAL, "        ---        ")
		self.add_centered_item(self.glass_value)
		
		self.add_descr_label(parent, "Magnification factor")
		self.mag_factor_list = [2, 4, 5, 8, 10, 20]
		self.mag_factor_cb = wx.ComboBox(parent, WID.ANY, size=(80,-1), style=wx.CB_READONLY)
		self.mag_factor_cb.AppendItems(["x %i"%f for f in self.mag_factor_list])
		self.mag_factor_cb.Select(3)
		self.mag_factor_cb.Bind(wx.EVT_COMBOBOX, self.on_change_mag_factor)
		self.add_centered_item(self.mag_factor_cb)
	#}}}
	
	def reset(self):
		pass

	def SetGlassImage(self, pil_image):#{{{
		if pil_image is None:
			self.glass.bitmap = None
		else:
			nx, ny = self.glass.GetSize()
			img = wx.EmptyImage(nx, ny)
			rpil = pil_image.resize((nx, ny))
			img.SetData(rpil.convert("RGB").tostring())
			img.SetAlphaData(rpil.tostring()[3::4])
			self.glass.bitmap = img.ConvertToBitmap()
		self.glass.Refresh()
	#}}}

	def RefreshPhysicalValue(self, label):#{{{
		"""
		"""
		self.glass_value.SetLabel(label)
	#}}}

	def on_change_mag_factor(self, event):#{{{
		self.set_mag_factor(event.GetSelection())
	#}}}
	
	def set_mag_factor(self, index):#{{{
		mgf = self.mag_factor_list[index]
		nx, ny = self.glass.GetSize()
		GenericTabMapWindow.mag_box_size = nx / mgf
		MagGlass.mag_factor = mgf
	#}}}
#}}}

__all__ = ["ParametersExpander", "LineOfSightExpander", "MagGlassExpander"]
